#!/usr/bin/python
# QLBR Charm
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ConfigParser import (
    NoOptionError,
    SafeConfigParser,
)
import os
import sys

from base64 import b64decode
from subprocess import check_call

from charmhelpers.core import (
    hookenv,
    host,
)
from charmhelpers import fetch
from charmhelpers.fetch.archiveurl import ArchiveUrlFetchHandler
from charmhelpers.fetch.bzrurl import BzrUrlFetchHandler


hooks = hookenv.Hooks()
log = hookenv.log
config = hookenv.config()

SERVICE = 'gunicorn'
MANAGE = os.path.join(config['qlbr_dir'], 'manage.py')
MANAGE_COMMAND = ['sudo', '-u', config['wsgi_user'], MANAGE]
DB_FILE = os.path.join(config['qlbr_dir'], 'qlbr', 'database_settings.py')
CACHE_FILE = os.path.join(config['qlbr_dir'], 'qlbr', 'cache_settings.py')


def _get_template(name):
    # Shamelessly yanked from uci-engine wsgi-app charm
    from jinja2 import (
        Environment,
        FileSystemLoader,
    )
    template_dir = os.path.join(hookenv.charm_dir(), 'templates')
    template_env = Environment(loader=FileSystemLoader(template_dir),
                               keep_trailing_newline=True)
    return template_env.get_template(name)


def _setup_cron():
    credentials = config['credentials_file']
    cron_file_path = config['cron_file']
    with open(cron_file_path, 'w') as cron_file:
        cron_file.write(
            _get_template('cron.tmpl').render(
                manage=MANAGE,
                credentials=credentials,
                user=config['wsgi_user'],
            )
        )
    os.chmod(cron_file_path, 0644)


def _write_credentials_file():
    file_name = config['credentials_file']
    credentials = b64decode(config['credentials'])
    with open(file_name, 'w') as credentials_file:
        credentials_file.write(credentials)
    parser = SafeConfigParser()
    parser.read(file_name)
    keys = (
        'consumer_key',
        'consumer_secret',
        'access_token',
        'access_secret',
    )
    for key in keys:
        try:
            parser.get('1', key)
        except NoOptionError:
            raise KeyError(
                '{} missing from credentials; cannot connect to launchpad'
                .format(key))


def _unpack_branch(branch=config['qlbr_branch'], qlbr_dir=config['qlbr_dir']):
    handler = BzrUrlFetchHandler()
    handler.branch(branch, qlbr_dir)


def _unpack_payload(payload=config['payload'], qlbr_dir=config['qlbr_dir']):
    if payload != '':
        handler = ArchiveUrlFetchHandler()
        handler.install(payload, qlbr_dir)


def _run_management_command(command):
    return check_call(MANAGE_COMMAND + command)


def _sync_db():
    _run_management_command(['syncdb', '--noinput'])
    _run_management_command(['drop_searches'])
    searches = os.path.join(config['qlbr_dir'], 'qlbr', 'searches.json')
    _run_management_command(['load_searches'], searches)


@hooks.hook('config-changed')
def config_changed():
    for key in config:
        if config.changed(key):
            log("config['{}'] changed from {} to {}".format(
                key, config.previous(key), config[key]))

    if config.changed('payload'):
        _unpack_payload()
    if config.changed('credentials'):
        _write_credentials_file()
    if config.changed('qlbr_branch'):
        _unpack_branch()
    config.save()


@hooks.hook('start')
def start():
    host.service_restart(SERVICE) or host.service_start(SERVICE)


@hooks.hook('stop')
def stop():
    host.service_stop(SERVICE)


@hooks.hook('install')
def install():
    wu = config['wsgi_user']
    wg = config['wsgi_group']
    qu = config['qlbr_user']
    qg = config['qlbr_group']
    host.adduser(wu)
    host.add_user_to_group(wu, wg)
    host.adduser(qu)
    host.add_user_to_group(qu, qg)

    for ppa in config['ppas'].split():
        fetch.add_source(ppa)
    fetch.apt_update()
    fetch.apt_install(config['distro_packages'].split())

    _unpack_payload()

    if not os.path.exists(MANAGE):
        _unpack_branch()

    _write_credentials_file()
    host.chownr(config['qlbr_dir'], qu, qg)
    host.mkdir(config['cache_dir'], wu, wg, 700)
    _setup_cron()


@hooks.hook('pgsql-relation-joined', 'pgsql-relation-changed')
def pgsql_relation_joined_changed():
    # Cribbed this check from the python-django charm
    database = hookenv.relation_get('database')
    if not database:
        return True

    user = hookenv.relation_get('user')
    password = hookenv.relation_get('password')
    host = hookenv.relation_get('host')
    port = 5432

    with open(DB_FILE, 'w') as settings_file:
        settings_file.write(
            _get_template('database_settings.py.tmpl').render(
                database=database,
                user=user,
                password=password,
                host=host,
                port=port,
                cachedir=config['cache_dir']
            )
        )

    _sync_db()


@hooks.hook('pgsql-relation-broken')
def pgsql_relation_broken():
    os.unlink(DB_FILE)


@hooks.hook('wsgi-file-relation-joined', 'wsgi-file-relation-changed')
def wsgi_file_relation_joined_changed():
    hookenv.relation_set(
        working_dir=config['qlbr_dir'],
        port=config['port'],
        wsgi_user=config['wsgi_user'],
        wsgi_group=config['wsgi_group'],
    )
    hookenv.open_port(config['port'])
    start()


@hooks.hook('wsgi-file-relation-broken')
def wsgi_file_relation_broken():
    hookenv.close_port(config['port'])
    stop()


@hooks.hook('memcached-relation-joined', 'memcached-relation-changed')
def memcached_relation_joined_changed():
    # Cribbed this check from the python-django charm
    host = hookenv.relation_get('host')
    if not host:
        return True

    port = hookenv.relation_get('port')

    with open(CACHE_FILE, 'w') as settings_file:
        settings_file.write(
            _get_template('cache_settings.py.tmpl').render(
                host=host,
                port=port,
            )
        )

    start()


@hooks.hook('memcached-relation-broken')
def memcached_relation_broken():
    os.unlink(CACHE_FILE)
    start()


if __name__ == "__main__":
    # execute a hook based on the name the program is called by
    hooks.execute(sys.argv)
